@@ -14,7 +14,6 @@
 #    under the License.
 
 import fixtures
-from keystoneclient import fixture as keystone_client_fixture
 import mock
 from requests_mock.contrib import fixture as requests_mock_fixture
 from six.moves.urllib import parse
@@ -22,10 +21,11 @@ from six.moves.urllib import parse
 from cinderclient import client
 from cinderclient import exceptions
 from cinderclient import shell
-from cinderclient.tests.unit.fixture_data import base as fixture_base
-from cinderclient.tests.unit.fixture_data import keystone_client
+from cinderclient.v2 import volumes
+from cinderclient.v2 import shell as test_shell
 from cinderclient.tests.unit import utils
 from cinderclient.tests.unit.v2 import fakes
+from cinderclient.tests.unit.fixture_data import keystone_client
 
 
 class ShellTest(utils.TestCase):
@@ -55,18 +55,16 @@ class ShellTest(utils.TestCase):
         self.requests = self.useFixture(requests_mock_fixture.Fixture())
         self.requests.register_uri(
             'GET', keystone_client.BASE_URL,
-            text=keystone_client.keystone_request_callback
-        )
-        token = keystone_client_fixture.V2Token()
-        s = token.add_service('volume', 'cinder')
-        s.add_endpoint(public='http://127.0.0.1:8776')
+            text=keystone_client.keystone_request_callback)
 
-        self.requests.post(keystone_client.BASE_URL + 'v2.0/tokens',
-                           json=token)
-        self.requests.get(
-            'http://127.0.0.1:8776',
-            json=fixture_base.generate_version_output()
-        )
+        self.cs = mock.Mock()
+
+    def _make_args(self, args):
+        class Args(object):
+            def __init__(self, entries):
+                self.__dict__.update(entries)
+
+        return Args(args)
 
     def tearDown(self):
         # For some methods like test_image_meta_bad_action we are
@@ -125,7 +123,8 @@ class ShellTest(utils.TestCase):
                                'snapshot_id': None,
                                'metadata': {'key1': '"--test1"'},
                                'volume_type': None,
-                               'description': None}}
+                               'description': None,
+                               'multiattach': False}}
         self.assert_called_anytime('POST', '/volumes', expected)
 
     def test_metadata_args_limiter_display_name(self):
@@ -145,7 +144,8 @@ class ShellTest(utils.TestCase):
                                'snapshot_id': None,
                                'metadata': {'key1': '"--t1"'},
                                'volume_type': None,
-                               'description': None}}
+                               'description': None,
+                               'multiattach': False}}
         self.assert_called_anytime('POST', '/volumes', expected)
 
     def test_delimit_metadata_args(self):
@@ -165,7 +165,8 @@ class ShellTest(utils.TestCase):
                                'metadata': {'key1': '"test1"',
                                             'key2': '"test2"'},
                                'volume_type': None,
-                               'description': None}}
+                               'description': None,
+                               'multiattach': False}}
         self.assert_called_anytime('POST', '/volumes', expected)
 
     def test_delimit_metadata_args_display_name(self):
@@ -185,7 +186,8 @@ class ShellTest(utils.TestCase):
                                'snapshot_id': None,
                                'metadata': {'key1': '"t1"'},
                                'volume_type': None,
-                               'description': None}}
+                               'description': None,
+                               'multiattach': False}}
         self.assert_called_anytime('POST', '/volumes', expected)
 
     def test_list_filter_status(self):
@@ -274,7 +276,8 @@ class ShellTest(utils.TestCase):
             with mock.patch('cinderclient.utils.print_list') as mock_print:
                 self.run_command(cmd)
                 mock_print.assert_called_once_with(
-                    mock.ANY, mock.ANY, sortby_index=None)
+                    mock.ANY, mock.ANY, exclude_unavailable=True,
+                    sortby_index=None)
 
     def test_list_reorder_without_sort(self):
         # sortby_index is 0 without sort information
@@ -282,7 +285,8 @@ class ShellTest(utils.TestCase):
             with mock.patch('cinderclient.utils.print_list') as mock_print:
                 self.run_command(cmd)
                 mock_print.assert_called_once_with(
-                    mock.ANY, mock.ANY, sortby_index=0)
+                    mock.ANY, mock.ANY, exclude_unavailable=True,
+                    sortby_index=0)
 
     def test_list_availability_zone(self):
         self.run_command('availability-zone-list')
@@ -334,6 +338,13 @@ class ShellTest(utils.TestCase):
     def test_create_size_required_if_not_snapshot_or_clone(self):
         self.assertRaises(SystemExit, self.run_command, 'create')
 
+    def test_create_size_zero_if_not_snapshot_or_clone(self):
+        expected = {'volume': {'status': 'creating',
+                               'size': 0}}
+        self.run_command('create 0')
+        self.assert_called_anytime('POST', '/volumes', partial_body=expected)
+        self.assert_called('GET', '/volumes/1234')
+
     def test_show(self):
         self.run_command('show 1234')
         self.assert_called('GET', '/volumes/1234')
@@ -361,10 +372,38 @@ class ShellTest(utils.TestCase):
         self.run_command('backup-create 1234 --incremental')
         self.assert_called('POST', '/backups')
 
+    def test_backup_force(self):
+        self.run_command('backup-create 1234 --force')
+        self.assert_called('POST', '/backups')
+
     def test_restore(self):
         self.run_command('backup-restore 1234')
         self.assert_called('POST', '/backups/1234/restore')
 
+    @mock.patch('cinderclient.utils.print_dict')
+    @mock.patch('cinderclient.utils.find_volume')
+    def test_do_backup_restore(self,
+                               mock_find_volume,
+                               mock_print_dict):
+        backup_id = '1234'
+        volume_id = '5678'
+        input = {
+            'backup': backup_id,
+            'volume': volume_id
+        }
+
+        args = self._make_args(input)
+        with mock.patch.object(self.cs.restores,
+                               'restore') as mocked_restore:
+            mock_find_volume.return_value = volumes.Volume(self,
+                                                           {'id': volume_id},
+                                                           loaded = True)
+            test_shell.do_backup_restore(self.cs, args)
+            mocked_restore.assert_called_once_with(
+                input['backup'],
+                volume_id)
+            self.assertTrue(mock_print_dict.called)
+
     def test_record_export(self):
         self.run_command('backup-export 1234')
         self.assert_called('GET', '/backups/1234/export_record')
@@ -457,6 +496,25 @@ class ShellTest(utils.TestCase):
         expected = {'os-reset_status': {'status': 'error'}}
         self.assert_called('POST', '/volumes/1234/action', body=expected)
 
+    def test_reset_state_with_attach_status(self):
+        self.run_command('reset-state --attach-status detached 1234')
+        expected = {'os-reset_status': {'status': 'available',
+                                        'attach_status': 'detached'}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
+    def test_reset_state_with_attach_status_with_flag(self):
+        self.run_command('reset-state --state in-use '
+                         '--attach-status attached 1234')
+        expected = {'os-reset_status': {'status': 'in-use',
+                                        'attach_status': 'attached'}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
+    def test_reset_state_with_reset_migration_status(self):
+        self.run_command('reset-state --reset-migration-status 1234')
+        expected = {'os-reset_status': {'status': 'available',
+                                        'migration_status': 'none'}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
     def test_reset_state_multiple(self):
         self.run_command('reset-state 1234 5678 --state error')
         expected = {'os-reset_status': {'status': 'error'}}
@@ -494,14 +552,32 @@ class ShellTest(utils.TestCase):
         self.assert_called_anytime('POST', '/snapshots/5678/action',
                                    body=expected)
 
+    def test_backup_reset_state(self):
+        self.run_command('backup-reset-state 1234')
+        expected = {'os-reset_status': {'status': 'available'}}
+        self.assert_called('POST', '/backups/1234/action', body=expected)
+
+    def test_backup_reset_state_with_flag(self):
+        self.run_command('backup-reset-state --state error 1234')
+        expected = {'os-reset_status': {'status': 'error'}}
+        self.assert_called('POST', '/backups/1234/action', body=expected)
+
+    def test_backup_reset_state_multiple(self):
+        self.run_command('backup-reset-state 1234 5678')
+        expected = {'os-reset_status': {'status': 'available'}}
+        self.assert_called_anytime('POST', '/backups/1234/action',
+                                   body=expected)
+        self.assert_called_anytime('POST', '/backups/5678/action',
+                                   body=expected)
+
     def test_type_list(self):
         self.run_command('type-list')
-        self.assert_called_anytime('GET', '/types')
-
-    def test_type_list_all(self):
-        self.run_command('type-list --all')
         self.assert_called_anytime('GET', '/types?is_public=None')
 
+    def test_type_show(self):
+        self.run_command('type-show 1')
+        self.assert_called('GET', '/types/1')
+
     def test_type_create(self):
         self.run_command('type-create test-type-1')
         self.assert_called('POST', '/types')
@@ -524,6 +600,15 @@ class ShellTest(utils.TestCase):
                          '--is-public=False')
         self.assert_called('POST', '/types', body=expected)
 
+    def test_type_update(self):
+        expected = {'volume_type': {'name': 'test-type-1',
+                                    'description': 'test_type-1-desc',
+                                    'is_public': False}}
+        self.run_command('type-update --name test-type-1 '
+                         '--description=test_type-1-desc '
+                         '--is-public=False 1')
+        self.assert_called('PUT', '/types/1', body=expected)
+
     def test_type_access_list(self):
         self.run_command('type-access-list --volume-type 3')
         self.assert_called('GET', '/types/3/os-volume-type-access')
@@ -535,6 +620,16 @@ class ShellTest(utils.TestCase):
         self.assert_called('POST', '/types/3/action',
                            body=expected)
 
+    def test_type_access_add_project_by_name(self):
+        expected = {'addProjectAccess': {'project': '101'}}
+        with mock.patch('cinderclient.utils.find_resource') as mock_find:
+            mock_find.return_value = '3'
+            self.run_command('type-access-add --volume-type type_name \
+                              --project-id 101')
+            mock_find.assert_called_once_with(mock.ANY, 'type_name')
+        self.assert_called('POST', '/types/3/action',
+                           body=expected)
+
     def test_type_access_remove_project(self):
         expected = {'removeProjectAccess': {'project': '101'}}
         self.run_command('type-access-remove '
@@ -552,7 +647,7 @@ class ShellTest(utils.TestCase):
         - one per volume type to retrieve the encryption type information
         """
         self.run_command('encryption-type-list')
-        self.assert_called_anytime('GET', '/types')
+        self.assert_called_anytime('GET', '/types?is_public=None')
         self.assert_called_anytime('GET', '/types/1/encryption')
         self.assert_called_anytime('GET', '/types/2/encryption')
 
@@ -592,8 +687,67 @@ class ShellTest(utils.TestCase):
         - one GET request to retrieve the relevant volume type information
         - one GET request to retrieve the relevant encryption type information
         - one PUT request to update the encryption type information
+        Verify that the PUT request correctly parses encryption-type-update
+        parameters from sys.argv
         """
-        self.skipTest("Not implemented")
+        parameters = {'--provider': 'EncryptionProvider', '--cipher': 'des',
+                      '--key-size': 1024, '--control-location': 'back-end'}
+
+        # Construct the argument string for the update call and the
+        # expected encryption-type body that should be produced by it
+        args = ' '.join(['%s %s' % (k, v) for k, v in parameters.items()])
+        expected = {'encryption': {'provider': 'EncryptionProvider',
+                                   'cipher': 'des',
+                                   'key_size': 1024,
+                                   'control_location': 'back-end'}}
+
+        self.run_command('encryption-type-update 1 %s' % args)
+        self.assert_called('GET', '/types/1/encryption')
+        self.assert_called_anytime('GET', '/types/1')
+        self.assert_called_anytime('PUT', '/types/1/encryption/provider',
+                                   body=expected)
+
+    def test_encryption_type_update_no_attributes(self):
+        """
+        Test encryption-type-update shell command.
+
+        Verify two GETs/one PUT requests are made per command invocation:
+        - one GET request to retrieve the relevant volume type information
+        - one GET request to retrieve the relevant encryption type information
+        - one PUT request to update the encryption type information
+        """
+        expected = {'encryption': {}}
+        self.run_command('encryption-type-update 1')
+        self.assert_called('GET', '/types/1/encryption')
+        self.assert_called_anytime('GET', '/types/1')
+        self.assert_called_anytime('PUT', '/types/1/encryption/provider',
+                                   body=expected)
+
+    def test_encryption_type_update_default_attributes(self):
+        """
+        Test encryption-type-update shell command.
+
+        Verify two GETs/one PUT requests are made per command invocation:
+        - one GET request to retrieve the relevant volume type information
+        - one GET request to retrieve the relevant encryption type information
+        - one PUT request to update the encryption type information
+        Verify that the encryption-type body produced contains default None
+        values for all specified parameters.
+        """
+        parameters = ['--cipher', '--key-size']
+
+        # Construct the argument string for the update call and the
+        # expected encryption-type body that should be produced by it
+        args = ' '.join(['%s' % (p) for p in parameters])
+        expected_pairs = [(k.strip('-').replace('-', '_'), None) for k in
+                          parameters]
+        expected = {'encryption': dict(expected_pairs)}
+
+        self.run_command('encryption-type-update 1 %s' % args)
+        self.assert_called('GET', '/types/1/encryption')
+        self.assert_called_anytime('GET', '/types/1')
+        self.assert_called_anytime('PUT', '/types/1/encryption/provider',
+                                   body=expected)
 
     def test_encryption_type_delete(self):
         """
@@ -608,17 +762,55 @@ class ShellTest(utils.TestCase):
         self.assert_called_anytime('GET', '/types/1')
 
     def test_migrate_volume(self):
-        self.run_command('migrate 1234 fakehost --force-host-copy=True')
+        self.run_command('migrate 1234 fakehost --force-host-copy=True '
+                         '--lock-volume=True')
         expected = {'os-migrate_volume': {'force_host_copy': 'True',
+                                          'lock_volume': 'True',
                                           'host': 'fakehost'}}
         self.assert_called('POST', '/volumes/1234/action', body=expected)
 
     def test_migrate_volume_bool_force(self):
-        self.run_command('migrate 1234 fakehost --force-host-copy')
+        self.run_command('migrate 1234 fakehost --force-host-copy '
+                         '--lock-volume')
         expected = {'os-migrate_volume': {'force_host_copy': True,
+                                          'lock_volume': True,
                                           'host': 'fakehost'}}
         self.assert_called('POST', '/volumes/1234/action', body=expected)
 
+    def test_migrate_volume_bool_force_false(self):
+        # Set both --force-host-copy and --lock-volume to False.
+        self.run_command('migrate 1234 fakehost --force-host-copy=False '
+                         '--lock-volume=False')
+        expected = {'os-migrate_volume': {'force_host_copy': 'False',
+                                          'lock_volume': 'False',
+                                          'host': 'fakehost'}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
+        # Do not set the values to --force-host-copy and --lock-volume.
+        self.run_command('migrate 1234 fakehost')
+        expected = {'os-migrate_volume': {'force_host_copy': False,
+                                          'lock_volume': False,
+                                          'host': 'fakehost'}}
+        self.assert_called('POST', '/volumes/1234/action',
+                           body=expected)
+
+    def test_replication_enable(self):
+        self.run_command('replication-enable 1234')
+        self.assert_called('POST', '/volumes/1234/action')
+
+    def test_replication_disable(self):
+        self.run_command('replication-disable 1234')
+        self.assert_called('POST', '/volumes/1234/action')
+
+    def test_replication_list_targets(self):
+        self.run_command('replication-list-targets 1234')
+        self.assert_called('POST', '/volumes/1234/action')
+
+    def test_replication_failover(self):
+        self.run_command('replication-failover 1234 target')
+        expected = {'os-failover_replication': {'secondary': 'target'}}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
+
     def test_snapshot_metadata_set(self):
         self.run_command('snapshot-metadata 1234 set key1=val1 key2=val2')
         self.assert_called('POST', '/snapshots/1234/metadata',
@@ -850,7 +1042,7 @@ class ShellTest(utils.TestCase):
                           self.run_command,
                           'consisgroup-update 1234')
 
-    def test_consistencygroup_create_from_src(self):
+    def test_consistencygroup_create_from_src_snap(self):
         self.run_command('consisgroup-create-from-src '
                          '--name cg '
                          '--cgsnapshot 1234')
@@ -861,14 +1053,101 @@ class ShellTest(utils.TestCase):
                 'description': None,
                 'user_id': None,
                 'project_id': None,
-                'status': 'creating'
+                'status': 'creating',
+                'source_cgid': None
             }
         }
         self.assert_called('POST', '/consistencygroups/create_from_src',
                            expected)
 
-    def test_consistencygroup_create_from_src_bad_request(self):
+    def test_consistencygroup_create_from_src_cg(self):
+        self.run_command('consisgroup-create-from-src '
+                         '--name cg '
+                         '--source-cg 1234')
+        expected = {
+            'consistencygroup-from-src': {
+                'name': 'cg',
+                'cgsnapshot_id': None,
+                'description': None,
+                'user_id': None,
+                'project_id': None,
+                'status': 'creating',
+                'source_cgid': '1234'
+            }
+        }
+        self.assert_called('POST', '/consistencygroups/create_from_src',
+                           expected)
+
+    def test_consistencygroup_create_from_src_fail_no_snap_cg(self):
         self.assertRaises(exceptions.BadRequest,
                           self.run_command,
                           'consisgroup-create-from-src '
                           '--name cg')
+
+    def test_consistencygroup_create_from_src_fail_both_snap_cg(self):
+        self.assertRaises(exceptions.BadRequest,
+                          self.run_command,
+                          'consisgroup-create-from-src '
+                          '--name cg '
+                          '--cgsnapshot 1234 '
+                          '--source-cg 5678')
+
+    def test_set_image_metadata(self):
+        self.run_command('image-metadata 1234 set key1=val1')
+        expected = {"os-set_image_metadata": {"metadata": {"key1": "val1"}}}
+        self.assert_called('POST', '/volumes/1234/action',
+                           body=expected)
+
+    def test_unset_image_metadata(self):
+        self.run_command('image-metadata 1234 unset key1')
+        expected = {"os-unset_image_metadata": {"key": "key1"}}
+        self.assert_called('POST', '/volumes/1234/action',
+                           body=expected)
+
+    def _get_params_from_stack(self, pos=-1):
+        method, url = self.shell.cs.client.callstack[pos][0:2]
+        path, query = parse.splitquery(url)
+        params = parse.parse_qs(query)
+        return path, params
+
+    def test_backup_list_all_tenants(self):
+        self.run_command('backup-list --all-tenants=1 --name=bc '
+                         '--status=available --volume-id=1234')
+        expected = {
+            'all_tenants': ['1'],
+            'name': ['bc'],
+            'status': ['available'],
+            'volume_id': ['1234'],
+        }
+
+        path, params = self._get_params_from_stack()
+
+        self.assertEqual('/backups/detail', path)
+        self.assertEqual(4, len(params))
+
+        for k in params.keys():
+            self.assertEqual(expected[k], params[k])
+
+    def test_backup_list_volume_id(self):
+        self.run_command('backup-list --volume-id=1234')
+        self.assert_called('GET', '/backups/detail?volume_id=1234')
+
+    def test_backup_list(self):
+        self.run_command('backup-list')
+        self.assert_called('GET', '/backups/detail')
+
+    def test_get_capabilities(self):
+        self.run_command('get-capabilities host')
+        self.assert_called('GET', '/capabilities/host')
+
+    def test_image_metadata_show(self):
+        # since the request is not actually sent to cinder API but is
+        # calling the method in :class:`v2.fakes.FakeHTTPClient` instead.
+        # Thus, ignore any exception which is false negative compare
+        # with real API call.
+        try:
+            self.run_command('image-metadata-show 1234')
+        except Exception:
+            pass
+        expected = {"os-show_image_metadata": None}
+        self.assert_called('POST', '/volumes/1234/action', body=expected)
